//Based on http://en.wikipedia.org/wiki/A*_search_algorithm

#include <deque>
#include <map>
#include <algorithm>
#include <GL/glut.h>
#include <vector>
#include <float.h>
#include <assert.h>
#include <time.h>
#include <stdlib.h>

#define ASSERT assert

inline float pow2(float x){return x*x;}


template<typename T,typename U>
bool isIn(const T &collection,U value)
{
	return collection.end()!=std::find(collection.begin(),collection.end(),value);
}

struct D3DXVECTOR3{
	float x,y,z;
};

#define NUM_MAX_ADJACENCY 16

struct GRAPH_NODE{
	D3DXVECTOR3 vPos;

	struct GRAPH_ADJ{
		int idx;
		float fDist;
	};
	GRAPH_ADJ adjArray[NUM_MAX_ADJACENCY];
	int numAdjs;
};

typedef std::deque<int> GRAPHLIST;
typedef std::map<int,float> GARPH_COST_MAP;

struct GRAPH_SORTER{
	GARPH_COST_MAP &_rMap;
	GRAPH_SORTER(GARPH_COST_MAP &rMap):_rMap(rMap){}
	bool operator()(const int &a,const int &b){
		float fA=_rMap[a];
		float fB=_rMap[b];
		return fA < fB;
	}
};

float getHeuristicCost(const GRAPH_NODE* a,const GRAPH_NODE* b)
{
	return sqrt(pow2(a->vPos.x-b->vPos.x) + pow2(a->vPos.y-b->vPos.y) + pow2(a->vPos.z-b->vPos.z));
}

float getActualCost(const GRAPH_NODE* a,const GRAPH_NODE* b)
{
	return sqrt(pow2(a->vPos.x-b->vPos.x) + pow2(a->vPos.y-b->vPos.y) + pow2(a->vPos.z-b->vPos.z));
}

bool astar_search(std::map<int,int> &parentOf,const GRAPH_NODE* pNodes,int nStart,int nGoal)
{
	//tB[h͋[sAJnn_AS[n_^B

	GRAPHLIST openList;
	GRAPHLIST closedList;
	GARPH_COST_MAP g_score;	// Distance from start along optimal path.
	GARPH_COST_MAP f_score;	// Estimated total distance from start to goal through y.

	const GRAPH_NODE* pStart=&pNodes[nStart];
	const GRAPH_NODE* pGoal=&pNodes[nGoal];

	g_score[nStart] = 0.0f;
	f_score[nStart] = getHeuristicCost(pStart, pGoal);

	//I[vXgɊJnm[hǉ
	openList.push_back(nStart);

	//I[vXgł͂Ȃ
	while(!openList.empty()){
		//dݏŃ\[g;łȕ擪ɗ
		std::sort(openList.begin(),openList.end(),GRAPH_SORTER(f_score));

		//݂̃m[h = I[vXg̍łȃm[h
		const int nCurrentNode=openList.front();
		const GRAPH_NODE* pCurrentNode = &pNodes[nCurrentNode];
		float fCostOfCurrentNode = g_score[nCurrentNode];

		//if ݂̃m[h = ړIñm[h then
		if(nCurrentNode == nGoal){
			//oH̊
			//pGoalparentOfHĂ΂悢(goal->start)
			//KvȂAĂяonGoalnStartւ鎖(start->goal)
			//ex. idx=parentOf[idx]
			return true;
		}else{
			//݂̃m[hN[YhXgɈڂ
			closedList.push_back(nCurrentNode);
			openList.pop_front();

			//אڃm[h
			//݂̗אڂem[h𒲂ׂ
			for(int i=0;i<pCurrentNode->numAdjs;i++){
				const int nNeigh=pCurrentNode->adjArray[i].idx;
				const GRAPH_NODE* pNeigh=&pNodes[nNeigh];
				float fCostActual=pCurrentNode->adjArray[i].fDist;
				if(isIn(closedList,nNeigh)){
					continue;
				}
				float costEstimated = fCostOfCurrentNode + fCostActual;

				bool bBetterFound=false;
				if(!isIn(openList,nNeigh)){
					//I[vXgɈڂ
					openList.push_back(nNeigh);
					bBetterFound=true;
				}else if(costEstimated < g_score[nNeigh]){
					bBetterFound=true;
				}
				if(bBetterFound){
					//eL^
					parentOf[nNeigh]=nCurrentNode;
					g_score[nNeigh] = costEstimated;
					f_score[nNeigh] = costEstimated + getHeuristicCost(pNeigh, pGoal);
				}
			}
		}
	}
	//Ts
	return false;
}



//---

std::vector<GRAPH_NODE> g_nodes;
std::map<int,int> g_routeMap;
int g_marked=-1;
int g_goalIdx=-1;
int g_startIdx=-1;

//w肵n_ł߂_T
//NbNɂnCCgɎgp
int findNearestNode(int x,int y)
{
	float fMin=FLT_MAX;
	int idx=-1;
	for(size_t i=0;i<g_nodes.size();i++){
		float dist=pow2(g_nodes[i].vPos.x - x) + pow2(g_nodes[i].vPos.y - y);
		if(dist<fMin){
			fMin=dist;
			idx=i;
		}
	}
	return idx;
}

//m[hԘA
void bindNodes(int _a,int _b)
{
	int indices[2]={_a,_b};
	for(int j=0;j<2;j++){
		int a=indices[j];
		int b=indices[1-j];

		bool bRemoved=false;
		for(int i=0;i<g_nodes[a].numAdjs;){
			if(g_nodes[a].adjArray[i].idx==b){
				//ɘAĂȂA폜
				bRemoved=true;
				g_nodes[a].numAdjs--;
				for(;i<g_nodes[a].numAdjs;i++){
					g_nodes[a].adjArray[i]=g_nodes[a].adjArray[i+1];
				}
			}else{
				i++;
			}
		}
		if(!bRemoved){
			//AȂ̂ŐVKA
			ASSERT(g_nodes[a].numAdjs<NUM_MAX_ADJACENCY-1);
			g_nodes[a].adjArray[g_nodes[a].numAdjs].idx=b;
			g_nodes[a].adjArray[g_nodes[a].numAdjs].fDist=getActualCost(&g_nodes[a],&g_nodes[b]);
			g_nodes[a].numAdjs++;
		}
	}
}

void makeRandomMap()
{
	g_nodes.clear();
	g_startIdx=-1;
	g_goalIdx=-1;
	g_marked=-1;
	g_routeMap.clear();

	int num=rand()%30+30;
	g_nodes.resize(num);
	for(int i=0;i<num;i++){
		g_nodes[i].vPos.x=rand()%800;
		g_nodes[i].vPos.y=rand()%600;
		g_nodes[i].vPos.z=0.0f;
		g_nodes[i].numAdjs=0;
	}

	for(int i=0;i<num;i++){
		int a=rand()%num;
		int b=rand()%num;
		if(g_nodes[a].numAdjs>=NUM_MAX_ADJACENCY-1 || g_nodes[b].numAdjs>=NUM_MAX_ADJACENCY-1)continue;
		bindNodes(a,b);
	}
}

void makeGridMap()
{
	g_nodes.clear();
	g_startIdx=-1;
	g_goalIdx=-1;
	g_marked=-1;
	g_routeMap.clear();

	int nx=40;
	int ny=40;
	int num=nx*ny;
	g_nodes.resize(num);
	for(int y=0;y<ny;y++){
		for(int x=0;x<nx;x++){
			int i=x+nx*y;
			g_nodes[i].vPos.x=x*15;
			g_nodes[i].vPos.y=y*15;
			g_nodes[i].vPos.z=0.0f;
			g_nodes[i].numAdjs=0;
		}
	}

	for(int y=0;y<ny;y++){
		for(int x=0;x<nx;x++){
			int a=x+nx*y;
			for(int i=0;i<9;i++){
				if(i==4)continue;
				int dx=i%3-1;
				int dy=i/3-1;
				if(x+dx<0 || x+dx>=nx || y+dy<0 || y+dy>=ny)continue;
				if(dx<0)continue;
				if(rand()%3!=0){
					int b=x+dx+nx*(y+dy);
					bindNodes(a,b);
				}
			}
		}
	}
}

void findPath()
{
	if(g_goalIdx==-1 || g_startIdx==-1)return;

	if(!astar_search(g_routeMap,&g_nodes[0],g_startIdx,g_goalIdx)){
		g_routeMap.clear();
	}
}

void mouse(int button , int state , int x , int y)
{
	if(button == GLUT_LEFT_BUTTON && state == GLUT_DOWN){
		int modifier=glutGetModifiers();
		if(modifier & GLUT_ACTIVE_ALT){
			int idx=findNearestNode(x,y);
			if(-1==idx){
				//_Ȃ
				return;
			}else{
				g_startIdx=idx;
			}
			findPath();
		}else if(modifier & GLUT_ACTIVE_SHIFT){
			int idx=findNearestNode(x,y);
			if(-1==idx){
				//_Ȃ
				return;
			}else{
				g_goalIdx=idx;
			}
			findPath();
		}else if(modifier &  GLUT_ACTIVE_CTRL ){
			int idx=findNearestNode(x,y);
			if(-1==idx){
				//_Ȃ
				return;
			}
			if(g_marked==-1){
				g_marked=idx;
			}else{
				bindNodes(g_marked,idx);
				g_marked=idx;
			}
		}else{
			//m[hǉ
			GRAPH_NODE node;
			node.vPos.x=x;
			node.vPos.y=y;
			node.vPos.z=0.0f;
			node.numAdjs=0;
			g_nodes.push_back(node);
		}
	}else if(button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN){
		g_marked=-1;
	}
	glutPostRedisplay();
}

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);

	glColor3f(1.0f,1.0f,1.0f);
	glPointSize(3.0f);
	glBegin(GL_POINTS);
	for(size_t i=0;i<g_nodes.size();i++){
		glVertex2f(g_nodes[i].vPos.x,g_nodes[i].vPos.y);
	}
	glEnd();

	glPointSize(5.0f);
	glBegin(GL_POINTS);
	if(g_marked!=-1){
		glColor3f(1.0f,0.0f,1.0f);
		glVertex2f(g_nodes[g_marked].vPos.x,g_nodes[g_marked].vPos.y);
	}
	if(g_startIdx!=-1){
		glColor3f(1.0f,0.0f,0.0f);
		glVertex2f(g_nodes[g_startIdx].vPos.x,g_nodes[g_startIdx].vPos.y);
	}
	if(g_goalIdx!=-1){
		glColor3f(0.0f,1.0f,0.0f);
		glVertex2f(g_nodes[g_goalIdx].vPos.x,g_nodes[g_goalIdx].vPos.y);
	}
	glEnd();

	glColor3f(1.0f,0.5f,0.0f);
	glBegin(GL_LINES);
	for(size_t i=0;i<g_nodes.size();i++){
		for(int j=0;j<g_nodes[i].numAdjs;j++){
			glVertex2f(g_nodes[i].vPos.x,g_nodes[i].vPos.y);
			int k=g_nodes[i].adjArray[j].idx;
			glVertex2f(g_nodes[k].vPos.x,g_nodes[k].vPos.y);
		}
	}
	glEnd();

	if(g_goalIdx!=-1 && g_startIdx!=-1 && !g_routeMap.empty()){
		glColor3f(0.0f,1.0f,1.0f);
		glBegin(GL_LINE_STRIP);
		int idx=g_goalIdx;
		for(;;){
			glVertex2f(g_nodes[idx].vPos.x,g_nodes[idx].vPos.y);
			if(idx==g_startIdx)break;
			idx=g_routeMap[idx];
		}
		glEnd();
	}

	glutSwapBuffers();
}

void resize(int w, int h);
void kbdFunc(unsigned char key , int x , int y);
void idle();

int main(int argc,char **argv)
{
	srand((unsigned)time(NULL));

	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowSize(800 , 600);
	glutCreateWindow("A* test");
	glutReshapeFunc(resize);

	glutKeyboardFunc(kbdFunc);
	glutIdleFunc(idle);
	glutDisplayFunc(display);
	glutMouseFunc(mouse);

	glClearColor(0.5f,0.5f,0.5f,1.0f);

	printf(
"g:\n"
"[L[]\n"
"1:\t_ȃ}bv𐶐\n"
"2:\tiq̃}bv𐶐\n"
"f:\t_ȓ_Ԃ̃[g𐶐\n"
"Q,ESC:\tI\n"
"\n"
"[}EX]\n"
"NbN:\t_\n"
"ENbN:\t}[N\n"
"CTRL+NbN:\tԋ߂_}[NAʂɃ}[Nς݂̒_΁AƘA\n"
"SHIFT+NbN:\tԋ߂_S[ɐݒ\n"
"ALT+NbN:\tԋ߂_X^[gɐݒ\n"
		);

	glutMainLoop();

	return 0;
}

void resize(int w, int h)
{
	/* EBhEŜr[|[gɂ */
	glViewport(0, 0, w, h);

	/* ϊs̏ */
	glLoadIdentity();

	/* XN[̍Wn}EX̍WnɈv */
	glOrtho(-0.5, (GLdouble)w - 0.5, (GLdouble)h - 0.5, -0.5, -1.0, 1.0);
}

void kbdFunc(unsigned char key , int x , int y)
{
	switch(key){
	case 'q':
	case 'Q':
	case 27: //ESC
		exit(0);
		break;
	case '1':
		//auto alignment
		makeRandomMap();
		break;
	case '2':
		makeGridMap();
		break;
	case 'f':
	case 'F':
		if(!g_nodes.empty()){
			g_startIdx=rand()%g_nodes.size();
			g_goalIdx=rand()%g_nodes.size();
			findPath();
		}
		break;
	}
}

void idle()
{
	glutPostRedisplay();
}


